import base64
from .base import (
    VulnersApiBase,
    Endpoint,
    String,
    Integer,
    Float,
    Dict,
    List,
    UUID,
    Boolean,
)
from .base import ResultSet

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping


def make_result_set(content, headers):
    return ResultSet.from_dataset(content, int(headers["X-Total-Count"]))


class MappingObject(Mapping):
    __slots__ = ("_api",)

    def __init__(self, api, data):
        self._api = api
        self.__dict__ = data

    def __len__(self):
        return len(self.__dict__)

    def __getitem__(self, k):
        return self.__dict__[k]

    def __iter__(self):
        return iter(self.__dict__)

    def __repr__(self):
        return repr(self.__dict__)


class Project(MappingObject):
    def update(self, **kwargs):
        kwargs.setdefault("name", self.name)
        kwargs.setdefault("license_id", self.license["id"])
        kwargs.setdefault("notification", self.notification)
        self.__dict__ = self._api.update_project(self._id, **kwargs).__dict__

    def delete(self):
        self._api.delete_project(self._id)

    def get_tasks(self, *args, **kwargs):
        return self._api.get_tasks(self._id, *args, **kwargs)

    def create_task(self, *args, **kwargs):
        return self._api.create_task(self._id, *args, **kwargs)

    def get_results(self, *args, **kwargs):
        return self._api.get_results(self._id, *args, **kwargs)

    def get_statistics(self, *args, **kwargs):
        return self._api.get_statistics(self._id, *args, **kwargs)

    def delete_result(self, *args, **kwargs):
        return self._api.delete_result(self._id, *args, **kwargs)


class Task(MappingObject):
    def update(self, **kwargs):
        kwargs.setdefault("name", self.name)
        kwargs.setdefault("networks", self.networks)
        kwargs.setdefault("schedule", self.schedule)
        kwargs.setdefault("enabled", self.enabled)
        self.__dict__ = self._api.update_task(
            self.project_id, self._id, **kwargs
        ).__dict__

    def delete(self):
        self._api.delete_task(self.project_id, self._id)

    def start_task(self):
        self.__dict__ = self._api.stark_task(self.project_id, self._id).__dict__


class VScannerApi(VulnersApiBase):
    ratelimit_key = "vscanner"

    get_licenses = Endpoint(
        method="get",
        url="/api/v3/useraction/licenseids",
        description="Get user's license ids.",
    )
    get_projects = Endpoint(
        method="get",
        url="/api/v3/proxy/vscanner/projects",
        description="Get existing projects.",
        params=[
            ("offset", Integer(default=0)),
            ("limit", Integer(default=50)),
        ],
        wrapper=lambda api, c: [Project(api, x) for x in c],
    )
    get_quota = Endpoint(
        method="get",
        url="/api/v3/proxy/vscanner/quota/{uuid:license_id|License id}",
        description="Get api quota left.",
    )
    create_project = Endpoint(
        method="post",
        url="/api/v3/proxy/vscanner/projects",
        description="Create new project.",
        params=[
            ("name", String(description="New project name")),
            ("license_id", UUID(description="User's license id")),
            (
                "notification",
                Dict(
                    description=(
                        "Use VScannerApi.Notification or VScannerApi.DisabledNotification helpers "
                        "to create notification object."
                    )
                ),
            ),
        ],
        wrapper=Project,
    )
    update_project = Endpoint(
        method="put",
        url="/api/v3/proxy/vscanner/projects/{uuid:project_id|Project id}",
        description="Update existing project.",
        params=[
            ("name", String(description="Project name")),
            ("license_id", UUID(description="User's license id")),
            (
                "notification",
                Dict(
                    description=(
                        "Use VScannerApi.Notification or VScannerApi.DisabledNotification helpers "
                        "to create notification object."
                    )
                ),
            ),
        ],
        wrapper=Project,
    )
    delete_project = Endpoint(
        method="delete",
        url="/api/v3/proxy/vscanner/projects/{uuid:project_id|Project id}",
        description="Delete existing project.",
    )
    get_tasks = Endpoint(
        method="get",
        url="/api/v3/proxy/vscanner/projects/{uuid:project_id|Project id}/tasks",
        description="Get project tasks",
        params=[("offset", Integer(default=0)), ("limit", Integer(default=50))],
        wrapper=lambda api, c: [Task(api, x) for x in c],
    )
    create_task = Endpoint(
        method="post",
        url="/api/v3/proxy/vscanner/projects/{uuid:project_id|Project id}/tasks",
        description="Create new task.",
        params=[
            ("name", String(description="Task name")),
            ("networks", List(description="List of networks (ip or domains)")),
            ("schedule", String(description="Crontab string")),
            ("enabled", Boolean(description="Enable/disable task")),
        ],
        wrapper=Task,
    )
    update_task = Endpoint(
        method="put",
        url="/api/v3/proxy/vscanner/projects/{uuid:project_id|Project id}/tasks/{uuid:task_id|Task id}",
        description="Update task.",
        params=[
            ("name", String(description="Task name")),
            ("networks", List(description="List of networks (ip or domains)")),
            ("schedule", String(description="Crontab string")),
            ("enabled", Boolean(description="Enable/disable task")),
        ],
        wrapper=Task,
    )
    start_task = Endpoint(
        method="post",
        url="/api/v3/proxy/vscanner/projects/{uuid:project_id|Project id}/tasks/{uuid:task_id|Task id}/start",
        description="Start task asap.",
        wrapper=Task,
    )
    delete_task = Endpoint(
        method="delete",
        url="/api/v3/proxy/vscanner/projects/{uuid:project_id|Project id}/tasks/{uuid:task_id|Task id}",
        description="Delete task.",
    )
    get_results = Endpoint(
        method="get",
        url="/api/v3/proxy/vscanner/projects/{uuid:project_id|Project id}/results",
        description="Get results.",
        params=[
            (
                "search",
                String(
                    required=False,
                    description="Search text occurrence in fields: name, ip or resolved.",
                ),
            ),
            (
                "ip_address",
                String(
                    required=False,
                    description="Comma-separated list of ip-addresses, networks. "
                    "Example: '91.218.85.0/24,91.218.86.99'",
                ),
            ),
            (
                "include_ports",
                String(
                    required=False,
                    description="Comma-separated list of ports. Example: '80,8080'.",
                ),
            ),
            (
                "exclude_ports",
                String(
                    required=False,
                    description="Comma-separated list of ports. Example: '80,8080'.",
                ),
            ),
            ("min_cvss", Float(required=False, description="Minimum CVSS value.")),
            ("max_cvss", Float(required=False, description="Maximum CVSS value.")),
            ("highlight", Boolean(default=False)),
            (
                "not_older_then",
                String(
                    required=False,
                    description=(
                        "Minimum published date. Example: '1d' not older than day, "
                        "'5h' not older than 5 hours, '100m' not older than 100 minutes, "
                        "'2021-07-21' published after the specified date, "
                        "'2021-07-21T12:00:00Z' published after the specified time."
                    ),
                ),
            ),
            (
                "sort",
                String(
                    required=False,
                    description=(
                        "Sort by field. Allowable values are 'ip', 'name', 'published', 'resolved',"
                        "'min_cvss', 'max_cvss'. To sort in descending order, use '-'."
                        "Default value is '-published'."
                    ),
                ),
            ),
            ("offset", Integer(default=0)),
            ("limit", Integer(default=50)),
        ],
        content_handler=make_result_set,
    )
    delete_result = Endpoint(
        method="delete",
        url="/api/v3/proxy/vscanner/projects/{uuid:project_id|Project id}/results/{uuid:result_id|Result id}",
        description="Delete result by id.",
    )
    get_statistics = Endpoint(
        method="get",
        url="/api/v3/proxy/vscanner/projects/{uuid:project_id|Project id}/statistics",
        description="Get statistics.",
        params=[
            (
                "agg",
                String(
                    required=True,
                    description=(
                        "Comma-separated list of aggregations: "
                        "total_hosts, vulnerable_hosts, unique_cve, min_max_cvss, "
                        "vulnerabilities_rank, vulnerable_hosts_rank"
                    ),
                ),
            ),
            (
                "not_older_then",
                String(
                    required=False,
                    description=(
                        "Minimum published date. Example: '1d' not older than day, "
                        "'5h' not older than 5 hours, '100m' not older than 100 minutes, "
                        "'2021-07-21' published after the specified date, "
                        "'2021-07-21T12:00:00Z' published after the specified time."
                    ),
                ),
            ),
        ],
    )

    def get_image_binary(self, image_uri, as_base64=False):
        result, headers = self._send_request(
            "get", "/vscanner/screen/" + image_uri, {}, {}, self.ratelimit_key, "binary"
        )
        if as_base64:
            return base64.b64encode(result)
        return result

    @staticmethod
    def Notification(period, emails=None, slack_webhooks=None):
        """
        Create notification project

        period: one of "disabled", "asap", "hourly" or "daily"
        emails: list of emails
        slack_webhooks: list of slack webhooks
        """
        if period not in ("disabled", "asap", "hourly", "daily"):
            raise ValueError(
                'period expected to be one of "disabled", "asap", "hourly" or "daily"'
            )
        return {
            "period": period,
            "email": emails or [],
            "slack": slack_webhooks or [],
        }

    @staticmethod
    def DisabledNotification():
        """
        Create stub notification object with "disabled" period and empty methods
        """
        return {
            "period": "disabled",
            "email": [],
            "slack": [],
        }



import warnings
import zipfile
import io
import json
from .base import VulnersApiBase, ParamError, validate_params
from .base import Endpoint, String, Integer, Dict, List, Tuple, Const, ResultSet, Boolean


class VulnersApi(VulnersApiBase):
    search_size = 100
    default_fields = (
        "id",
        "title",
        "description",
        "type",
        "bulletinFamily",
        "cvss",
        "published",
        "modified",
        "lastseen",
        "href",
        "sourceHref",
        "sourceData",
        "cvelist",
    )

    __search = Endpoint(
        method="post",
        url="/api/v3/search/lucene/",
        params=[
            ("query", String()),
            ("skip", Integer()),
            ("size", Integer()),
            ("fields", Tuple()),
        ],
    )

    @validate_params(
        query=String(),
        limit=Integer(minimum=1, maximum=100),
        offset=Integer(minimum=0, maximum=10000),
        fields=Tuple(item=String()),
    )
    def find(self, query, limit=20, offset=0, fields=default_fields):
        """
        Search in Vulners database.

        query: Vulners query. See https://vulners.com/help for the details.
        limit: The maximum number of documents to return.
        offset: Skip this amount of documents. 10000 is the hard limit.
        fields: List of fields to return.

        Returns list of the documents.
        Use .total to get the total number of found documents.
        """
        end = min(10000, offset + limit)
        if offset >= end:
            return ResultSet.from_dataset(
                [], self.__search(query, 0, 1, ["id"])["total"]
            )
        else:
            search = self.__search(query, offset, limit, fields)
            return ResultSet.from_dataset(
                [e["_source"] for e in search["search"]], search["total"]
            )

    @validate_params(
        query=String(),
        limit=Integer(minimum=1, maximum=10000),
        offset=Integer(minimum=0, maximum=10000),
        fields=Tuple(item=String()),
    )
    def find_all(self, query, limit=20, offset=0, fields=default_fields):
        """
        Search in Vulners database and returns up to 10000 documents.

        query: Vulners query. See https://vulners.com/help for the details.
        limit: The maximum number of documents to return..
        offset: The number of documents to skip in the result set. 10000 is the hard limit.
        fields: List of fields to return.

        Returns list of the documents.
        Use .total to get the total number of found documents.
        """
        end = min(10000, offset + limit)
        if offset >= end:
            return ResultSet.from_dataset(
                [], self.__search(query, 0, 1, ["id"])["total"]
            )
        result = ResultSet()
        batch_size = min(self.search_size, limit)
        for skip in range(offset, end, batch_size):
            chunk = self.__search(query, skip, min(batch_size, end - skip), fields)
            result += [e["_source"] for e in chunk["search"]]
            if len(chunk["search"]) < batch_size:
                result.total = chunk["total"]
                break
        return result

    @validate_params(
        query=String(),
        lookup_fields=Tuple(item=String()),
        limit=Integer(minimum=1, maximum=100),
        offset=Integer(minimum=0, maximum=10000),
        fields=Tuple(item=String()),
    )
    def find_exploit(
        self, query, lookup_fields=None, limit=20, offset=0, fields=default_fields
    ):
        """
        Search in Vulners database for the exploits.

        query: Software name and criteria.
        lookup_fields: Make a strict search using lookup fields. Like ["title"]
        limit: The maximum number of documents to return.
        offset: Skip this amount of documents. 10000 is the hard limit.
        fields: List of fields to return.

        Returns list of the documents.
        Use .total to get the total number of found documents.
        """
        if lookup_fields:
            search_query = "bulletinFamily:exploit AND (%s)" % (
                " OR ".join('%s:"%s"' % (field, query) for field in lookup_fields)
            )
        else:
            search_query = "bulletinFamily:exploit AND %s" % query
        return self.find(search_query, limit, offset, fields)

    @validate_params(
        query=String(),
        lookup_fields=Tuple(item=String()),
        limit=Integer(minimum=1, maximum=100),
        offset=Integer(minimum=0, maximum=10000),
        fields=Tuple(item=String()),
    )
    def find_exploit_all(
        self, query, lookup_fields=None, limit=20, offset=0, fields=default_fields
    ):
        """
        Search in Vulners database for the exploits and returns up to 10000 documents.

        query: Software name and criteria.
        lookup_fields: Make a strict search using lookup fields. Like ["title"]
        limit: The maximum number of documents to return.
        offset: Skip this amount of documents. 10000 is the hard limit.
        fields: List of fields to return.

        Returns list of the documents.
        Use .total to get the total number of found documents.
        """
        if lookup_fields:
            search_query = "bulletinFamily:exploit AND (%s)" % (
                " OR ".join('%s:"%s"' % (field, query) for field in lookup_fields)
            )
        else:
            search_query = "bulletinFamily:exploit AND %s" % query
        return self.find_all(search_query, limit, offset, fields)

    get_web_application_rules = Endpoint(method="get", url="/api/v3/burp/rules/")

    def _get_burp_software_content(content, _):
        result = {}
        # noinspection PyUnresolvedReferences
        for elem in content.get("search", ()):
            elem = elem["_source"]
            result.setdefault(elem["bulletinFamily"], []).append(elem)
        return result

    __get_burp_software = Endpoint(
        method="post",
        url="/api/v3/burp/software/",
        params=[("software", String()), ("version", String()), ("type", String()), ("exactmatch", Boolean(default=False))],
        content_handler=_get_burp_software_content,
    )

    del _get_burp_software_content

    @validate_params(name=String(), version=String())
    def get_software_vulnerabilities(self, name, version):
        """
        Find software vulnerabilities using name and version.

        name: Software name, e.g. 'httpd'
        version: Software version, e.g. '2.1'
        """
        return self.__get_burp_software(name, version, "software")

    @validate_params(cpe=String())
    def get_cpe_vulnerabilities(self, cpe, exactmatch=False):
        """
        Find software vulnerabilities using CPE string. See CPE references at https://cpe.mitre.org/specification/

        cpe: CPE software string, see https://cpe.mitre.org/specification/
        exactmatch:  if true searches only for bulletins corresponding to the specified minor version and revision
        """
        cpe_split = cpe.split(":")
        if len(cpe_split) <= 4:
            raise ParamError("Malformed %s", "cpe")
        if cpe_split[1] == '2.3':
            version = cpe_split[5]
        elif cpe_split[1] in '/a/o/h':
            version = cpe_split[4]
        else:
            raise ParamError("Malformed %s", "cpe")

        return self.__get_burp_software(cpe, version, "cpe", exactmatch=exactmatch)

    get_multiple_bulletins = Endpoint(
        method="post",
        url="/api/v3/search/id/",
        description="Fetch multiple bulletins by ids.",
        params=[
            (
                "id",
                Tuple(
                    item=String(), description="List of ID's. E.g., ['CVE-2017-14174']"
                ),
            ),
            ("fields", Tuple(item=String(), default=default_fields)),
        ],
        content_handler=lambda c, _: c["documents"],
    )

    @validate_params(id=String(), fields=Tuple(item=String()))
    def get_bulletin(self, id, fields=default_fields):
        """
        Fetch bulletin by id.

        id: Bulletin ID. E.g., "CVE-2017-14174"
        """
        return self.get_multiple_bulletins([id], fields).get(id, {})

    get_multiple_bulletin_references = Endpoint(
        method="post",
        url="/api/v3/search/id/",
        description="",
        params=[
            (
                "id",
                List(
                    item=String(), description="List of ID's. E.g., ['CVE-2017-14174']"
                ),
            ),
            ("fields", Tuple(item=String(), default=default_fields)),
            ("references", Const(True)),
        ],
        content_handler=lambda c, _: c["references"],
    )

    @validate_params(id=String(), fields=Tuple(item=String()))
    def get_bulletin_references(self, id, fields=default_fields):
        """
        Fetch bulletin references by identificator

        identificator: Bulletin ID. E.g., "CVE-2017-14174"
        """
        return self.get_multiple_bulletin_references([id], fields=fields).get(id, {})

    @validate_params(kbid=String())
    def get_kb_seeds(self, kbid):
        """
        Returns superseeds and parentseeds for the given KB.
        Superseeds means "KB which are covered by this KB".
        Parentseeds means "KB which are covering this KB".

        superseeds --> KB --> parentseeds

        kbid: Microsoft KB identificator
        """
        candidate = self.get_bulletin(id=kbid, fields=["superseeds", "parentseeds"])
        return {
            "superseeds": candidate.get("superseeds", []),
            "parentseeds": candidate.get("parentseeds", []),
        }

    @validate_params(kbid=String(), fields=Tuple(item=String()))
    def get_kb_updates(self, kbid, fields=default_fields):
        """
        Returns list of updates for KB.

        kbid: Microsoft KB identificator.
        """
        query = "type:msupdate AND kb:(%s)" % kbid
        return self.find_all(query, limit=1000, fields=fields)

    os_audit = Endpoint(
        method="post",
        url="/api/v3/audit/audit/",
        description=(
            "Linux Audit API for analyzing package vulnerabilities.\n"
            "Accepts RPM and DEB based package lists.\n"
            "For collecting RPM use command: rpm -qa --qf '%{NAME}-%{VERSION}-%{RELEASE}.%{ARCH}\\\\n'\n"
            "For collecting DEB use command: dpkg-query -W -f='${Package} ${Version} ${Architecture}\\\\n'\n"
        ),
        params=[
            (
                "os",
                String(
                    description="Full name of the OS. Like Ubuntu, Debian, rhel, oraclelinux"
                ),
            ),
            ("version", String(description="OS version")),
            (
                "packages",
                List(
                    item=String(),
                    description="List of the installed packages",
                    param="package",
                ),
            ),
        ],
    )

    software_audit = Endpoint(
        method="post",
        url="/api/v3/burp/packages/",
        description=(
            "Software audit allow you to analyse software name / version pairs for the CVE's.\n"
            "Packages input format, list of dicts:\n"
            "[{'software': 'Mozilla Firefox', 'version': '80.0.1'}]\n"
        ),
        params=[
            (
                "os",
                String(
                    description="Full name of the OS. Like Ubuntu, Debian, rhel, oraclelinux"
                ),
            ),
            ("version", String(description="OS version", param="osVersion")),
            ("packages", List(item=Dict(), description="List of the software dicts")),
        ],
    )

    kb_audit = Endpoint(
        method="post",
        url="/api/v3/audit/kb/",
        description="Windows KB audit function",
        params=[
            (
                "os",
                String(description="Windows OS name, like 'Windows Server 2012 R2'"),
            ),
            (
                "kb_list",
                List(
                    item=String(),
                    description="List of installed KB's, ['KB2918614', 'KB2918616']",
                    param="kbList",
                ),
            ),
        ],
    )

    get_suggestion = Endpoint(
        method="post",
        url="/api/v3/search/suggest/",
        params=[("type", Const("distinct")), ("field_name", String(param="fieldName"))],
        content_handler=lambda c, _: c["suggest"],
    )

    get_ai_score = Endpoint(
        method="post",
        url="/api/v3/ai/scoretext/",
        params=[
            ("text", String()),
        ],
        content_handler=lambda c, _: c.get("score", 0),
    )

    query_autocomplete = Endpoint(
        method="post",
        url="/api/v3/search/autocomplete/",
        description="Ask Vulners for possible suggestions to complete your query",
        params=[("query", String(description="Vulners Search query"))],
        content_handler=lambda c, _: [q[0] for q in c["suggestions"]],
    )

    # noinspection PyTypeChecker
    def _unpack_json_file(c, _):
        with zipfile.ZipFile(io.BytesIO(c)) as zip_file:
            if len(zip_file.namelist()) > 1:
                raise Exception("Unexpected file count in Vulners ZIP archive")
            file_name = zip_file.namelist()[0]
            return json.loads(zip_file.open(file_name).read())

    __archive_collection = Endpoint(
        method="get",
        url="/api/v3/archive/collection/",
        params=[("type", String()), ("datefrom", String()), ("dateto", String())],
        result_type="zipjson",
        content_handler=_unpack_json_file,
    )

    __distributive = Endpoint(
        method="get",
        url="/api/v3/archive/distributive/",
        params=[("os", String()), ("version", String())],
        result_type="zipjson",
        content_handler=_unpack_json_file,
    )

    del _unpack_json_file

    @validate_params(collection=String(), start_date=String(), end_date=String())
    def get_collection(
        self, collection, start_date="1950-01-01", end_date="2199-01-01"
    ):
        """
        Get entire collection data

        :param collection: Collection name
        """
        collections = self.get_suggestion("type")
        if collection not in collections:
            raise ParamError(
                "Unknown %%s. Available values are %s" % (collections,), "collection"
            )
        return self.__archive_collection(
            type=collection, datefrom=start_date, dateto=end_date
        )

    @validate_params(os=String(), version=String())
    def get_distributive(self, os, version):
        """
        Get dict with data for OS vulnerability assessment

        os: OS name
        version: OS version
        """
        supported_os = self.get_suggestion("affectedPackage.OS")
        if os.lower() not in [os_name.lower() for os_name in supported_os]:
            raise ParamError(
                "Unknown %%s. Available values are %s" % (supported_os,), "os"
            )
        data = self.__distributive(os=os, version=version)
        return [bulletin["_source"] for bulletin in data]


_Unset = object()


# noinspection PyPep8Naming
class DeprecatedVulnersApi(VulnersApi):
    def __init__(self, *args, **kwargs):
        super(DeprecatedVulnersApi, self).__init__(*args, **kwargs)
        warnings.warn(
            "Vulners is deprecated and will be removed in future release. "
            "Use VulnersApi instead.",
            DeprecationWarning,
        )

    def search(self, query, limit=100, offset=0, fields=None):
        warnings.warn(
            "search() is deprecated and will be removed in future release. "
            "Use VulnersApi.find_all() instead.",
            DeprecationWarning,
        )
        return self.find_all(
            query, limit=limit, offset=offset, fields=fields or self.default_fields
        )

    def searchPage(self, query, pageSize=20, offset=0, fields=None):
        warnings.warn(
            "searchPage() is deprecated and will be removed in future release. "
            "Use VulnersApi.find() instead.",
            DeprecationWarning,
        )
        return self.find(
            query, limit=pageSize, offset=offset, fields=fields or self.default_fields
        )

    def searchExploit(
        self, query, lookup_fields=None, limit=100, offset=0, fields=None
    ):
        warnings.warn(
            "searchExploit() is deprecated and will be removed in future release. "
            "Use VulnersApi.find_exploit_all() instead.",
            DeprecationWarning,
        )
        lookup_fields = lookup_fields or ()
        if isinstance(lookup_fields, (set, list)):
            lookup_fields = tuple(lookup_fields)
        return self.find_exploit_all(
            query,
            lookup_fields=lookup_fields,
            limit=limit,
            offset=offset,
            fields=fields or self.default_fields,
        )

    def searchExploitPage(
        self, query, lookup_fields=None, limit=100, offset=0, fields=None
    ):
        warnings.warn(
            "searchExploitPage() is deprecated and will be removed in future release. "
            "Use VulnersApi.find_exploit() instead.",
            DeprecationWarning,
        )
        lookup_fields = lookup_fields or ()
        if isinstance(lookup_fields, (set, list)):
            lookup_fields = tuple(lookup_fields)
        return self.find_exploit(
            query,
            lookup_fields=lookup_fields,
            limit=limit,
            offset=offset,
            fields=fields or self.default_fields,
        )

    def softwareVulnerabilities(self, name, version, maxVulnerabilities=_Unset):
        warnings.warn(
            "softwareVulnerabilities() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_software_vulnerabilities() instead.",
            DeprecationWarning,
        )
        if maxVulnerabilities is not _Unset:
            warnings.warn(
                "maxVulnerabilities is deprecated and will be removed in future release.",
                DeprecationWarning,
            )
        return self.get_software_vulnerabilities(name, version)

    def cpeVulnerabilities(self, cpeString, maxVulnerabilities=_Unset):
        warnings.warn(
            "cpeVulnerabilities() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_cpe_vulnerabilities() instead.",
            DeprecationWarning,
        )
        if maxVulnerabilities is not _Unset:
            warnings.warn(
                "maxVulnerabilities is deprecated and will be removed in future release.",
                DeprecationWarning,
            )
        return self.get_cpe_vulnerabilities(cpeString)

    def audit(self, os, os_version, package):
        warnings.warn(
            "audit() is deprecated and will be removed in future release. "
            "Use VulnersApi.os_audit() instead.",
            DeprecationWarning,
        )
        return self.os_audit(os, os_version, package)

    def kbAudit(self, os, kb_list):
        warnings.warn(
            "kbAudit() is deprecated and will be removed in future release. "
            "Use VulnersApi.kb_audit() instead.",
            DeprecationWarning,
        )
        return self.kb_audit(os, kb_list)

    def document(self, identificator, fields=None):
        warnings.warn(
            "document() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_bulletin() instead.",
            DeprecationWarning,
        )
        return self.get_bulletin(identificator, fields=fields or self.default_fields)

    def documentList(self, identificatorList, fields=None):
        warnings.warn(
            "documentList() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_multiple_bulletins() instead.",
            DeprecationWarning,
        )
        return self.get_multiple_bulletins(
            identificatorList, fields=fields or self.default_fields
        )

    def references(self, identificator, fields=None):
        warnings.warn(
            "references() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_bulletin_references() instead.",
            DeprecationWarning,
        )
        return self.get_bulletin_references(
            identificator, fields=fields or self.default_fields
        )

    def referencesList(self, identificatorList, fields=None):
        warnings.warn(
            "referencesList() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_multiple_bulletin_references() instead.",
            DeprecationWarning,
        )
        return self.get_multiple_bulletin_references(
            identificatorList, fields=fields or self.default_fields
        )

    def collections(self):
        warnings.warn(
            "collections() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_suggestion('type') instead.",
            DeprecationWarning,
        )
        return self.get_suggestion("type")

    def suggest(self, field_name):
        warnings.warn(
            "suggest() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_suggestion() instead.",
            DeprecationWarning,
        )
        return self.get_suggestion(field_name)

    def aiScore(self, text):
        warnings.warn(
            "aiScore() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_ai_score() instead.",
            DeprecationWarning,
        )
        return self.get_ai_score(text)

    def rules(self):
        warnings.warn(
            "rules() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_web_application_rules() instead.",
            DeprecationWarning,
        )
        return self.get_web_application_rules()

    def kbSuperseeds(self, kb_identificator):
        warnings.warn(
            "kbSuperseeds() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_kb_seeds() instead.",
            DeprecationWarning,
        )
        return self.get_kb_seeds(kb_identificator)

    def kbUpdates(self, kb_identificator, fields=None):
        warnings.warn(
            "kbUpdates() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_kb_updates() instead.",
            DeprecationWarning,
        )
        return self.get_kb_updates(
            kb_identificator, fields=fields or self.default_fields
        )

    def autocomplete(self, query):
        warnings.warn(
            "autocomplete() is deprecated and will be removed in future release. "
            "Use VulnersApi.query_autocomplete() instead.",
            DeprecationWarning,
        )
        return self.query_autocomplete(query)

    def archive(self, collection, start_date="1950-01-01", end_date="2199-01-01"):
        warnings.warn(
            "archive() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_collection() instead.",
            DeprecationWarning,
        )
        return self.get_collection(collection, start_date, end_date)

    def distributive(self, os, version):
        warnings.warn(
            "distributive() is deprecated and will be removed in future release. "
            "Use VulnersApi.get_distributive() instead.",
            DeprecationWarning,
        )
        return self.get_distributive(os, version)


Vulners = DeprecatedVulnersApi



# -*- coding: utf-8 -*-

__version__ = "2.0.2"

import warnings
from .vulners import VulnersApi, Vulners
from .vscanner import VScannerApi
from .base import VulnersApiError


warnings.simplefilter("always", DeprecationWarning)



import os
import stat
import re
import uuid
import requests
import threading
import json
import functools
import inspect
from time import time, sleep
from collections import defaultdict, OrderedDict
from six import string_types, with_metaclass
from requests import adapters, cookies
from urllib3.util.retry import Retry
from appdirs import user_cache_dir
from . import __version__


try:
    _getargspec = inspect.getfullargspec  # noqa
except AttributeError:
    _getargspec = inspect.getargspec  # noqa


class VulnersApiError(Exception):
    def __init__(self, http_status, data):
        super(VulnersApiError, self).__init__(data)
        self.http_status = http_status


class VulnersApiMeta(type):
    def __new__(cls, name, bases, attrs):
        new_type = type.__new__
        if not bases:
            return new_type(cls, name, bases, attrs)
        for key, value in list(attrs.items()):
            if isinstance(value, Endpoint):
                attrs[key] = value.build(key, attrs.get("ratelimit_key"))
        new_class = new_type(cls, name, bases, attrs)
        return new_class


class RateLimitBucket(object):
    """An implementation of the Token Bucket algorithm."""

    def __init__(self, rate=10.0, burst=1.0):
        """Initialise an instance of the RateLimit allowing a default rate of 10 calls
        per 1 seconds when no arguments are supplied.
        :param rate:
            The number of calls per second. Default 10.
        """
        self._rate = float(rate)
        self._burst = min(float(burst), self._rate)
        self._allowance = self._burst
        self._last_check = time()

    def update(self, rate, burst=1.0):
        self._rate = float(rate)
        self._burst = min(float(burst), self._rate)

    def consume(self):
        while 1:
            now = time()
            # number of seconds since the last call
            delta = now - self._last_check
            self._last_check = now
            # increase the number of allowed calls
            self._allowance += delta * self._rate
            if self._allowance > self._burst:
                # don't allow more than "burst" calls
                self._allowance = self._burst
            if self._allowance < 1:
                # cold down
                sleep((1 - self._allowance) / self._rate)
                continue
            self._allowance -= 1
            break


class PersistentCookieJar(cookies.RequestsCookieJar):

    _cookie_locks = defaultdict(threading.RLock)

    def __init__(self, file_path=None, *args, **kwargs):
        super(PersistentCookieJar, self).__init__(*args, **kwargs)
        self._set_cookie_counter = 0
        if file_path:
            self._file_path = file_path
        else:
            cache_dir = user_cache_dir("Vulners", "Vulners.Inc")
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir)
            self._file_path = os.path.join(cache_dir, "cookies.txt")
        self._recover()

    def dump(self):
        if self._set_cookie_counter:
            with self._cookie_locks[self._file_path]:
                with open(self._file_path, "wb") as cookie_file:
                    cookie_file.write(
                        json.dumps(self.get_dict(), ensure_ascii=True).encode("ascii")
                    )
            self._set_cookie_counter = 0

    def _recover(self):
        try:
            mode = os.stat(self._file_path).st_mode
        except OSError:
            return
        if not stat.S_ISREG(mode):
            raise OSError("%s is not a regular file" % self._file_path)
        with self._cookie_locks[self._file_path]:
            with open(self._file_path, "rb") as cookie_file:
                try:
                    cookies_dict = json.loads(cookie_file.read())
                except (TypeError, ValueError):
                    return
        if not isinstance(cookies_dict, dict):
            return
            # this method internally calls set_cookie
        self.update(cookies_dict)
        self._set_cookie_counter = 0

    def extract_cookies(self, response, request):
        super(PersistentCookieJar, self).extract_cookies(response, request)
        self.dump()

    def set_cookie(self, *args, **kwargs):
        self._set_cookie_counter += 1
        return super(PersistentCookieJar, self).set_cookie(*args, **kwargs)


class VulnersApiBase(with_metaclass(VulnersApiMeta)):
    # Retry status codes
    retry_codes = (500, 502, 503, 504)
    # How many times to retry
    retry_count = 3

    _ratelimits = defaultdict(RateLimitBucket)

    def __init__(
        self, api_key, proxies=None, persistent=True, server_url="https://vulners.com"
    ):
        """
        Create VScanner API object.

        :param api_key:
            string with Vulners API key. You can obtain one from the https://vulners.com
        :param proxies:
            dict for proxy supporting. Example: {"https": "myproxy.com:3128"}
        """
        if not api_key:
            raise ValueError(
                "API key must be provided. You can obtain one for free at https://vulners.com"
            )

        if not isinstance(api_key, string_types):
            raise TypeError("api_key parameter must be a string value")

        self._sess = self._create_session(proxies, server_url, persistent)
        self._api_key = api_key
        self._server_url = server_url

    def _create_session(self, proxies, server_url, persistent):
        assert proxies is None or isinstance(proxies, dict), "proxies must be a dict"
        session = requests.session()
        retries = Retry(total=self.retry_count, status_forcelist=self.retry_codes)
        adapter = adapters.HTTPAdapter(max_retries=retries)
        session.mount(server_url, adapter)
        session.headers.update(
            {
                "Connection": "Keep-Alive",
                "User-Agent": "Vulners Python API %s" % __version__,
            }
        )
        if persistent:
            session.cookies = PersistentCookieJar()
        if proxies:
            session.proxies.update(proxies)
        return session

    @staticmethod
    def adapt_response(response, method, expected_result):
        if expected_result == "json":
            if response.content:
                result = response.json()
            else:
                result = None
            if response.status_code >= 400:
                raise VulnersApiError(response.status_code, result)
            if isinstance(result, dict) and "data" in result:
                data = result["data"]
                if isinstance(data, dict) and data.get("error"):
                    raise VulnersApiError(response.status_code, data)
                return data
            return result
        elif response.status_code >= 400:
            if response.content:
                result = response.json()
            else:
                result = None
            raise VulnersApiError(response.status_code, result)
        else:
            return response.content

    @staticmethod
    def _update_ratelimit(bucket, response):
        headers = response.headers
        limit = headers.get("X-Vulners-Ratelimit-Reqlimit")
        if limit:
            try:
                limit = float(limit) / 60.0
            except (TypeError, ValueError):
                return
            bucket.update(rate=limit)

    def _send_request(
        self,
        method,
        url,
        body_params,
        path_params,
        ratelimit_key=None,
        result_type="json",
    ):
        body_params["apiKey"] = self._api_key
        url = self._server_url + url.format(**path_params)
        kwargs = {"params" if method in ("get", "delete") else "json": body_params}
        bucket = self._ratelimits[ratelimit_key or url]
        bucket.consume()
        response = getattr(self._sess, method)(url, **kwargs)
        self._update_ratelimit(bucket, response)
        result = self.adapt_response(response, method, result_type)
        response.raise_for_status()
        return result, response.headers


_Nothing = object()


class ParamError(ValueError):
    def __init__(self, msg, param):
        self.msg = msg
        self.param = param

    def __str__(self):
        return self.msg % self.param


class Param(object):
    def __init__(self, default=_Nothing, description=None, required=True, param=None):
        if not required:
            if default is _Nothing:
                default = None
        self.param = param
        self.default = default
        self.required = required
        self.description = description


class Const(object):
    def __init__(self, value):
        self.value = value


class String(Param):
    def __init__(self, choices=None, *args, **kwargs):
        super(String, self).__init__(*args, **kwargs)
        self.choices = choices

    def validate(self, param, value):
        if not isinstance(value, string_types):
            raise ParamError("%s expected to be a string", param)
        if self.choices is not None and value not in self.choices:
            raise ParamError(
                "%%s expected to be on of (%s)" % ", ".join(self.choices), param
            )
        return value


class UUID(Param):
    @staticmethod
    def validate(param, value):
        try:
            return str(uuid.UUID(value))
        except (TypeError, ValueError):
            raise ParamError("%s expected to be an uuid", param)


class Dict(Param):
    @staticmethod
    def validate(param, value):
        if not isinstance(value, dict):
            raise ParamError("%s expected to be a dict", param)
        return value


class List(Param):
    def __init__(self, item=None, **kwargs):
        super(List, self).__init__(**kwargs)
        self.item = item

    @staticmethod
    def validate_items(validator, param, value):
        result = []
        for idx, val in enumerate(value):
            try:
                result.append(validator(idx, val))
            except ParamError as e:
                raise ParamError(e.msg, "%s.%s" % (param, e.param))
        return result

    def validate(self, param, value):
        if not isinstance(value, list):
            raise ParamError("%s expected to be a list", param)
        if self.item is not None:
            value = self.validate_items(self.item.validate, param, value)
        return value


class Tuple(List):
    def __init__(self, accept_list=True, **kwargs):
        super(Tuple, self).__init__(**kwargs)
        self.accept_list = accept_list

    def validate(self, param, value):
        if not isinstance(value, (list, tuple) if self.accept_list else tuple):
            raise ParamError("%s expected to be a tuple", param)
        if self.item is not None:
            value = self.validate_items(self.item.validate, param, value)
        return tuple(value)


class Integer(Param):
    def __init__(self, minimum=None, maximum=None, **kwargs):
        super(Integer, self).__init__(**kwargs)
        self.minimum = minimum
        self.maximum = maximum

    def validate(self, param, value):
        if not isinstance(value, int):
            raise ParamError("%s expected to be an int", param)
        if self.minimum is not None and value < self.minimum:
            raise ParamError("%%s must be greater or equal to %d" % self.minimum, param)
        if self.maximum is not None and value > self.maximum:
            raise ParamError("%%s must be less or equal to %d" % self.maximum, param)
        return value


class Float(Param):
    def __init__(self, minimum=None, maximum=None, **kwargs):
        super(Float, self).__init__(**kwargs)
        self.minimum = minimum
        self.maximum = maximum

    def validate(self, param, value):
        if not isinstance(value, float):
            raise TypeError("%s expected to be a float" % param)
        if self.minimum is not None and value < self.minimum:
            raise ParamError("%%s must be greater or equal to %d" % self.minimum, param)
        if self.maximum is not None and value > self.maximum:
            raise ParamError("%%s must be less or equal to %d" % self.maximum, param)
        return value


class Boolean(Param):
    @staticmethod
    def validate(param, value):
        if not isinstance(value, bool):
            raise TypeError("%s expected to be a bool" % param)
        return value


_unset = object()


def validate_params(**params):
    def decorator(func):
        spec = _getargspec(func)
        for param in params:
            if param not in spec.args:
                raise TypeError("No such argument %s" % param)
        spec_defaults = spec.defaults or ()
        default_values = (_unset,) * (
            len(spec.args) - len(spec_defaults)
        ) + spec_defaults
        defaults = {k: v for k, v in zip(spec.args, default_values) if v is not _unset}
        args = ", ".join([arg for arg in spec.args])
        func_args = ", ".join(
            [
                (name if default is _unset else ("%s=%r" % (name, default)))
                for name, default in zip(spec.args, default_values)
            ]
        )
        body = "\n  ".join(
            [
                (
                    ("if {var} is not _D[{var!r}]: " if var in defaults else "")
                    + "{var} = _V[{var!r}].validate({var!r}, {var})"
                ).format(var=var)
                for var in params
            ]
        )
        code = (
            "def {name}({func_args}):\n" "  {body}\n" "  return _func({args})"
        ).format(name=func.__name__, func_args=func_args, body=body, args=args)
        exec_locals = {"_V": params, "_D": defaults, "_func": func}
        exec(code, exec_locals, exec_locals)
        new_func = exec_locals[func.__name__]
        functools.update_wrapper(new_func, func)
        return new_func

    return decorator


class Endpoint(object):
    _mapping = {
        "str": String,
        "uuid": UUID,
        "int": Integer,
        "float": Float,
        "bool": Boolean,
        "dict": Dict,
    }

    def __call__(self, *args, **kwargs):
        raise RuntimeError("Only for typing")

    def __init__(
        self,
        method,
        url,
        description=None,
        params=None,
        result_type="json",
        content_handler=None,
        wrapper=None
    ):
        assert method in ("get", "post", "put", "delete")
        assert isinstance(url, string_types)
        assert description is None or isinstance(description, string_types)
        assert params is None or isinstance(params, list)
        self.path_params = []
        self.params = OrderedDict()
        for param in [item.strip("{}") for item in re.findall("{[^}]*}", url)]:
            if ":" in param:
                param_type, param_name = param.split(":", 1)
            else:
                param_type, param_name = "str", param
            if "|" in param_name:
                param_name, param_description = param_name.split("|", 1)
            else:
                param_description = ""
            url = url.replace("{" + param + "}", "{" + param_name + "}")
            param_obj = self._mapping[param_type](description=param_description)
            self.path_params.append(param_name)
            self.params[param_name] = param_obj
        self.method = method
        self.url = url
        self.description = description
        self.result_type = result_type
        self.content_handler = content_handler
        self.wrapper = wrapper
        if params:
            for k, v in params:
                self.params[k] = v

    def build(self, name, ratelimit_key=None):
        func_args = []
        func_args_with_default = []
        body_params = []
        path_params = []
        func_locals = {
            "_content_handler": self.content_handler,
            "_wrapper": self.wrapper
        }
        func_locals_reverse = {}
        func_doc = [self.description or name, ""]
        for key, param in self.params.items():
            if isinstance(param, Const):
                const_key = "_const%d" % len(func_locals)
                func_locals[const_key] = param.value
                body_params.append(
                    "body_params[{key!r}] = {const}".format(key=key, const=const_key)
                )
                continue
            if param.default is not _Nothing:
                default_key = "_default%d" % len(func_locals)
                func_locals[default_key] = param.default
                func_args_with_default.append("%s=%s" % (key, default_key))
            else:
                func_args.append(key)
            validate_key = "_validate%d" % len(func_locals)
            if param.validate in func_locals_reverse:
                validate_key = func_locals_reverse[param.validate]
            else:
                func_locals[validate_key] = param.validate
                func_locals_reverse[param.validate] = validate_key
            if key in self.path_params:
                path_params.append(
                    "{key!r}: {validator}({key!r}, {key})".format(
                        key=key, validator=validate_key
                    )
                )
            else:
                param_name = param.param or key
                if param.required:
                    body_params.append(
                        (
                            "body_params[{param_name!r}] = {validator}({key!r}, {key})"
                        ).format(param_name=param_name, key=key, validator=validate_key)
                    )
                else:
                    body_params.append(
                        (
                            "if {key} is not None: "
                            "body_params[{param_name!r}] = {validator}({key!r}, {key})"
                        ).format(param_name=param_name, key=key, validator=validate_key)
                    )
            if param.description:
                func_doc.append("%s: %s" % (key, param.description or ""))
        func_args += func_args_with_default
        code = (
            "def {name}(self, {func_args}):\n"
            "  '''{func_doc}'''\n"
            "  body_params = {{}}\n"
            "  {body_params}\n"
            "  path_params = {{{path_params}}}\n" 
            "  r = self._send_request({method!r}, {url!r}, body_params, path_params, {ratelimit_key!r}, {result!r})\n"
        ).format(
            name=name,
            func_args=", ".join(func_args),
            func_doc="\n".join(func_doc),
            method=self.method,
            url=self.url,
            body_params="\n  ".join(body_params),
            path_params=", ".join(path_params),
            ratelimit_key=ratelimit_key,
            result=self.result_type,
        )
        if self.content_handler:
            code += "  r = _content_handler(*r)\n"
        else:
            code += "  r = r[0]\n"
        if self.wrapper:
            code += "  r = _wrapper(self, r)\n"
        code += "  return r"
        exec(code, func_locals, func_locals)
        return func_locals[name]


class ResultSet(list):
    total = None

    @classmethod
    def from_dataset(cls, data, total):
        ret = cls(data)
        ret.total = total
        return ret



# -*- coding: utf-8 -*-

import re
import os
from setuptools import setup, find_packages


def get_version(package):
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def get_long_description(long_description_file):
    long_description = open(long_description_file).read()
    return long_description


setup(
    name="vulners",
    version=get_version("vulners"),
    description="Python library and command-line utility for Vulners (https://vulners.com)",
    long_description=get_long_description("README.md"),
    long_description_content_type="text/markdown",
    author="Kirill Ermakov, Andrei Churin",
    author_email="isox@vulners.com, aachurin@gmail.com",
    url="https://github.com/vulnersCom/api",
    packages=find_packages(exclude=["samples"]),
    install_requires=["requests", "six", "appdirs"],
    keywords=["security", "network", "vulners", "vulnerability", "CVE"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Software Development :: Version Control",
        "Topic :: Utilities",
    ],
)
